'''
@author: Sebastian Sudholt
@organization: TU Dortmund University, Dortmund, Germany
@contact: sebastian.sudholt@tu-dortmund.de
'''

import os
import time
import argparse
import numpy as np
from scipy.spatial.distance import pdist, squareform

from manifold.mds import BCMDS
from manifold.isomap import BCIsomap

from word_spotting_tools.word_spotting_io import load_word_images, spatial_pyramid_parser
from word_spotting_tools.spatial_pyramid import extract_spatial_pyramid
from word_spotting_tools.evaluation import map_from_pairwise_distances
from word_spotting_tools.bag_of_features import quantize_descriptors, create_visual_vocabulary


def main(gw_folder, codebook_size, descriptors_for_clustering, 
         step_size, descriptor_size, contrast_threshold, 
         spatial_pyramid_partitions, embedding_target_dim, 
         bc_isomap_k_neighbors, precomputed_codebook_path, 
         precomputed_codebook_data_type):
    '''
    This function runs the BC-MDS and BC-Isomap experiments on the George 
    Washington as presented in [1]. If you find this work or parts of it
    helpful, please cite us.
    
    Run "python george_washington.py -h" to learn about the parameters.
    
    [1]
    Sebastian Sudholt and Gernot Fink, A Modified Isomap Approach to Manifold
    Learning in Word Spotting, In Proc. of the 37th German Conference on
    Pattern Recognition, 2015, LNCS
    http://patrec.cs.tu-dortmund.de/pubs/abstracts/Sudholt2015-AMI.htm
    '''
    
    grid_offst = (descriptor_size/2 + 1, descriptor_size/2 +1)
    ground_truth_dir = os.path.join(gw_folder, 'ground_truth')
    pages_dir = os.path.join(gw_folder, 'pages')
    print('--- BC-MDS and BC-Isomap Experiment on the George Washington Database ---\n')
    
    print('---   PARAMETERS   ---')
    print('descriptor size: ' + str(descriptor_size))
    print('step size: ' + str(step_size))
    print('codebook size:' + str(codebook_size))
    print('descriptors for clustering: ' + str(descriptors_for_clustering))
    print('contrast threshold: ' + str(contrast_threshold))
    print('spatial pyramid: ' + str(spatial_pyramid_partitions))
    print('embedding target dimension: ' + str(embedding_target_dim))
    print('BC-Isomap k neighbors: ' + str(bc_isomap_k_neighbors))
    print('precomputed codebook: ' + str(precomputed_codebook_path))
    print('----------------------')
    
    #--- PART 1
    # extract the word images
    print('Extracting Word Images and Descriptors:')
    word_list = load_word_images(pages_dir=pages_dir, 
                                 ground_truth_dir=ground_truth_dir, 
                                 descriptor_size=descriptor_size,
                                 step_size=step_size, grid_offset=grid_offst, 
                                 contrast_threshold=contrast_threshold)
    
    #--- PART 2
    # create BoF representation
    #
    # 1) cluster the descriptors or use precomputed codebook
    if precomputed_codebook_path is not None:
        print('Loading visual vocabulary codebook... '),
        codebook = np.fromfile(precomputed_codebook_path, 
                               dtype=precomputed_codebook_data_type)
        codebook = np.reshape(codebook, (codebook_size, -1))
        print('[ Done ]')
    else:        
        codebook = create_visual_vocabulary(word_list, codebook_size, 
                                            descriptors_for_clustering)
    
    # 2) quantize the descriptors
    quantized_descriptors_list = quantize_descriptors(word_list, codebook)
    
    #--- PART 3
    # extract the Spatial Pyramids
    print('Extracting Spatial Pyramids... '),
    spatial_pyramids = [extract_spatial_pyramid(quantized_descriptors=quantized_descriptors, normalized_keypoint_locations=word[1], 
                                                partitions=spatial_pyramid_partitions, visual_vocabulary_size=codebook_size)
                        for quantized_descriptors, word in zip(quantized_descriptors_list, word_list)]
    spatial_pyramids = np.vstack(spatial_pyramids)
    print('[ Done ]')
    
    #--- PART 4
    # embed the data
    print('Embedding the Spatial Pyramids:')
    bc_mds = BCMDS(target_dim=embedding_target_dim, verbose=False)
    bc_isomap = BCIsomap(target_dim=embedding_target_dim, k_neighbors=bc_isomap_k_neighbors, verbose=False)
    
    print('\tBC-MDS... '),
    start_time = time.time()    
    mds_embedding = bc_mds.fit_transform(X=spatial_pyramids)    
    print('[ Done ] (took %f mins.)' % ((time.time()-start_time)/60.0))
    
    print('\tBC-Isomap... '),
    start_time = time.time() 
    iso_embedding = bc_isomap.fit_transform(X=spatial_pyramids)
    print('[ Done ] (took %f mins.)' % ((time.time()-start_time)/60.0))
    
    #--- PART 5
    # evaluate mAP
    print('Evaluating mAP (this might take a while)... '),
    pairwise_dist_baseline = squareform(pdist(spatial_pyramids, metric='braycurtis'))
    pairwise_dist_mds = squareform(pdist(mds_embedding, metric='euclidean'))
    pairwise_dist_iso = squareform(pdist(iso_embedding, metric='euclidean'))
    map_baseline = map_from_pairwise_distances(pairwise_dist_baseline, word_list)
    map_mds = map_from_pairwise_distances(pairwise_dist_mds, word_list)
    map_iso = map_from_pairwise_distances(pairwise_dist_iso, word_list)
    print('[ Done ]')
    
    #############################################################
    print('---------------------------------------------\nResults:')
    print('mAP Baseline: %f' % (map_baseline))
    print('mAP BC-MDS: %f' % (map_mds))
    print('mAP BC-Isomap: %f' % (map_iso))     
    
if __name__ == '__main__':
    help_gw_folder = 'Absolute path to the top directory of the George Washington Database'
    help_codebook_size = 'the amount of visual words'
    help_descriptor_size = 'Standard SIFT descriptors are extracted at a single \
                            scale with uniform orientation in our method. Their \
                            width and height are specified by this value.'
    help_dfc =  'As the creation of the visual vocabulary can be very time \
                 consuming, we only extract this many descriptors from the \
                 overall set and use them as representatives in the clustering.' 
    help_step_size = 'Descriptors are extracted for every word image in a dense \
                      grid with the step size in pixels specified by this value.' 
    help_ct = 'We only extract SIFT descriptors in the areas where significant \
               gradients exist. This value specifies the gradient magnitude \
               threshold.'
    help_td = 'This value specifies the target dimension BC-MDS and BC-Isomap \
               should project the spatial pyramids to.'
    help_knn = 'The BC-Isomap method needs to compute k neighbors to construct \
                the neighborhood graph.'
    help_cbp = 'The user can decide whether he wants to use a precomputed \
                codebook for the clustering or cluster a new codebook from \
                scratch. If you want to start from scratch simply set this \
                variable to None.'
    help_cbt = 'If you supply a codebook it is expected to be a binary file of \
                size n x 128 where n is the number of visual words. In order to \
                load the codebook, you need to specify the used data type here.'
    help_spp = 'We use a spatial pyramid as holistic word image representation. \
                This tuple of tuples specifies the layout of the spatial pyramid. \
                Each sub-tuple specifies the grid layout of an individual level. \
                (1,1) would thus stand for a single bin for the entire word image \
                (standard BoF histogram) while (2,3) would subdivide the word \
                image into two vertical and three horizontal slices at this level.'
                
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--gw_folder', '-gwf', action='store', type=str, required=True, help=help_gw_folder)
    parser.add_argument('--codebook_size', '-cbs', action='store', type=int, default=4096, help=help_codebook_size)
    parser.add_argument('--descriptors_for_clustering', '-dfc', action='store', type=int, default=600000, help=help_dfc)
    parser.add_argument('--step_size', '-ss', action='store', type=int, default=5, help=help_step_size)    
    parser.add_argument('--descriptor_size', '-ds', action='store', type=int, default=40, help=help_descriptor_size)
    parser.add_argument('--contrast_threshold', '-ct', action='store', type=float, default=0.01, help=help_ct)
    parser.add_argument('--embedding_target_dim', '-td', action='store', type=int, default=50, help=help_td)
    parser.add_argument('--bc_isomap_k_neighbors', '-knn', action='store', type=int, default=700, help=help_knn)
    parser.add_argument('--precomputed_codebook_path', '-pcbp', action='store', type=str, default='codebook_gw_4096x128_float32.bin', help=help_cbp)
    parser.add_argument('--precomputed_codebook_data_type', '-pcbt', action='store', type=type, default=np.float32, help=help_cbt)
    parser.add_argument('--spatial_pyramid_partitions', '-spp', action='store', type=spatial_pyramid_parser, default='((1,1),(1,2))', help=help_spp)
    
    args = parser.parse_args()
    main(**vars(args))