'''
@author: Sebastian Sudholt
@organization: TU Dortmund University, Dortmund, Germany
@contact: sebastian.sudholt@tu-dortmund.de
'''

import numpy as np

from scipy.sparse.csr import csr_matrix
from sklearn.utils.graph_shortest_path import graph_shortest_path
from scipy.spatial.distance import cdist, pdist, squareform


class BCIsomap(object):
    '''
    classdocs
    '''


    def __init__(self, target_dim, k_neighbors, verbose):
        '''
        Constructor
        @param target_dim: (int) the desired dimensionality after reduction
        @param k_neighbors: (int) the number of neighbors to use in the nn graph
        @param verbosity_level: (int) how talkative should I be
        '''
        self.target_dim = target_dim
        self.k_neighbors = k_neighbors
        self.verbose = verbose
        self.X = None
        self.L_sharp = None
        self.squared_geodesic_dists = None
        self.dist_offset = 0
        
    def fit_transform(self, X):
        '''
        Fits the Isomap model and returns the embedding for the training data
        
        @param X: (ndarray) training data for fitting the model of shape n x d (n samples, each of size d)
        '''
        self.X = X
        # find nearest neighbors and distances
        if self.verbose:
            print '[ Isomap ] computing %i nearest neighbors... ' % self.k_neighbors,        
        
        dist_mat = pdist(self.X, metric='braycurtis')
        dist_mat = squareform(dist_mat)
            
        # scale the distances
        self.dist_offset = np.sort(dist_mat, axis=None)[dist_mat.shape[0]] #extracts the smallest non zero entry from the dist matrix
        dist_mat -= self.dist_offset
        dist_mat += self.dist_offset * np.identity(dist_mat.shape[0])
        
        
        nn_inds = np.argsort(dist_mat, axis=1)
        not_connected_inds = nn_inds[:,self.k_neighbors+1:]
        for sample_idx, not_connected in enumerate(not_connected_inds):
            dist_mat[sample_idx][not_connected] = 0        
        sparse_dists = csr_matrix(dist_mat)
        if self.verbose:
            print('[ Done ]')
        
        # compute geodesic distances
        if self.verbose:
            print('[ Isomap ] computing geodesic distances... '),
        geodesic_dists = graph_shortest_path(sparse_dists, 
                                             method='auto',
                                             directed=False)        
        if self.verbose:
            print('[ Done ]')
            print('[ Isomap ] embedding:')
        
        # double center matrix
        self.squared_geodesic_dists = np.square(geodesic_dists, dtype=np.float64)
        self.data_mean = np.mean(self.squared_geodesic_dists, axis=1).flatten()
        H = np.identity(self.squared_geodesic_dists.shape[0], dtype='float64') - np.ones(self.squared_geodesic_dists.shape, dtype='float64')/self.squared_geodesic_dists.shape[0]
        B = -0.5 * np.dot(np.dot(H, self.squared_geodesic_dists), H)
        
        # extract eigenvalues
        eig_vals, eig_vecs = np.linalg.eig(B)
        
        # select the index for the largest target_dim eigen values
        # np.argsorts sorts in ascending order and [::-1] flips the
        # resulting array thus giving descending order
        inds = np.argsort(eig_vals)[::-1][:self.target_dim]
        eig_vals = eig_vals[inds]
        eig_vecs = eig_vecs[:,inds]
        
        # embed the data
        embedding = eig_vecs * np.sqrt(eig_vals)
        self.L_sharp = eig_vecs / np.sqrt(eig_vals)
        return embedding
    
    def transform(self, Y):
        '''
        Out-of-sample embedding for the supplied data with respect to the
        previously fitted model
        
        @param Y: (ndarray) n x d matrix to be embedded (n samples, each having d dimensions)
        '''
        if self.verbose:
            print('[ Isomap ] embedding out of sample, calculating geodesic distances... '),
        dists = cdist(Y, self.X, metric=self.metric) - self.dist_offset
        squared_dists = np.square(dists)
        inds = np.argsort(squared_dists, axis=1)[:,:self.k_neighbors]
        if self.verbose:
            print('[ Done ]')
        
        # compute geodesic distancess
        geo_sq_dists = np.zeros((Y.shape[0], self.X.shape[0]))
        for idx, nn_inds in zip(xrange(inds.shape[0]), inds):
            if self.verbose:
                print '\t[ %i / %i ] embedding sample... ' % (idx+1, Y.shape[0]),
            squared_nn_dists = squared_dists[idx][nn_inds]
            all_dists = self.squared_geodesic_dists[nn_inds] + np.transpose(np.atleast_2d(squared_nn_dists))
            geo_sq_dists[idx] = np.min(all_dists, axis=0)
            if self.verbose:
                print '[ Done ]'
            
        ###
        #embed
        embedding = 0.5 * np.dot((self.data_mean - geo_sq_dists), self.L_sharp) 
        return embedding
            
