'''
@author: Sebastian Sudholt
@organization: TU Dortmund University, Dortmund, Germany
@contact: sebastian.sudholt@tu-dortmund.de
'''

import numpy as np
from scipy.spatial.distance import cdist, pdist, squareform 

class BCMDS(object):
    def __init__(self, target_dim, verbose):
        self.target_dim = target_dim
        self.verbose = verbose
        self.X = None
        self.L_sharp = None
        self.data_mean = None
        self.metric = None
        
    def fit_transform(self, X):
        '''
        Fits the BC-MDS model and returns the embedding for the training data
        
        @param X: (ndarray) training data for fitting the model of shape n x d (n samples, each of size d)
        '''        
        # compute pairwise distances
        self.X = X
        dist_mat = pdist(X=self.X, metric='braycurtis')
        dist_mat = squareform(X=dist_mat)
        
        # double center matrix
        S = np.square(dist_mat, dtype=np.float64)
        self.data_mean = np.mean(S, axis=1).flatten()
        H = np.identity(S.shape[0], dtype='float64') - np.ones(S.shape, dtype='float64')/S.shape[0]
        B = -0.5 * np.dot(np.dot(H, S), H)
        
        # extract eigenvalues
        eig_vals, eig_vecs = np.linalg.eig(B)
        
        # select the index for the largest target_dim eigen values
        # np.argsorts sorts in ascending order and [::-1] flips the
        # resulting array thus giving descending order
        inds = np.argsort(eig_vals)[::-1][:self.target_dim]
        eig_vals = eig_vals[inds]
        eig_vecs = eig_vecs[:,inds]
        
        # embed the data
        embedding = eig_vecs * np.sqrt(eig_vals)
        self.L_sharp = eig_vecs / np.sqrt(eig_vals)
        return embedding
    
    def transform(self, Y):
        '''
        Out-of-sample embedding for the supplied data with respect to the
        previously fitted model
        
        @param Y: (ndarray) n x d matrix to be embedded (n samples, each having d dimensions)
        '''
        dists = cdist(Y, self.X, metric='braycurtis')
        squared_dists = np.square(dists)
        embedding = 0.5 * np.dot((self.data_mean - squared_dists), self.L_sharp)
        return embedding