'''
@author: Sebastian Sudholt, Leonard Rothacker
@organization: TU Dortmund University, Dortmund, Germany
@contact: sebastian.sudholt@tu-dortmund.de
'''

import numpy as np
import matplotlib.image as mpimg
import _vlfeat

def dense_features(image_filepath, descriptor_size, step_size, grid_bounds=None, grid_offset=None, contrast_threshold=0.0,adapt_size=False, pad_image=False, verbose=False):
    '''
    Compute dense SIFT descriptors
    
    '''
    img = mpimg.imread(image_filepath)
    img = normalize_image(img, verbose)  
    if pad_image:
        padded_image = np.zeros((img.shape[0] + descriptor_size, img.shape[1] + descriptor_size), dtype=np.float32)
        padded_image[descriptor_size/2:descriptor_size/2 + img.shape[0], descriptor_size/2:descriptor_size/2 + img.shape[1]] = img
        img = padded_image
    if adapt_size and ((min(img.shape[0], img.shape[1])-1) < descriptor_size or descriptor_size == 0):
        #if adapt_size is true the descriptor size is set the
        # to min(size, img.shape[0], img.shape[1]) so the maximum possible
        #descriptor size is used up to the specified size
        
        descriptor_size = (min(img.shape[0], img.shape[1])-1)
        print 'WARNING! descriptor size for image %s set to %i' % (image_filepath, descriptor_size)
        grid_offset = (descriptor_size/2+1, descriptor_size/2+1)
    
    if grid_offset is None:
        grid_offset = (3.0 / 2.0 * (descriptor_size/4), 3.0 / 2.0 * (descriptor_size/4))
    
    cell_offset = tuple(round(g_off - 3.0 / 2.0 * (descriptor_size/4)) for g_off in grid_offset) 
        
    if grid_bounds == None:
        # set bounds in terms of (x_min,y_min,x_max,y_max)
        grid_bounds = (0, 0, img.shape[1], img.shape[0])
    
    # Convert grid_bounds from upper-lower corner (x_min,y_min,x_max,y_max)
    # to matrix (row, column) format
    grid_bounds = (grid_bounds[1] + cell_offset[1], grid_bounds[0] + cell_offset[0], \
                   grid_bounds[3] - cell_offset[1], grid_bounds[2] - cell_offset[0])
    
    if any(c < 0 for c in grid_bounds) or grid_bounds[2] > img.shape[0] or grid_bounds[3] > img.shape[1] or\
            grid_bounds[0] >= grid_bounds[2] or grid_bounds[1] >= grid_bounds[3]:
        raise ValueError('grid_bounds out of image bounds or undefined grid area')
        
    # Image size needs to be at least twice the grid offset (in each dim): 
    # It is added / subtracted both (two times) at the beginning and end 
    # of the image (also see ValueError message text)
    if any(img.shape[i] < grid_offset[i] * 2 for i in range(2)):
        raise ValueError('image size too small for dense grid spec: no frames would be computed!')    
    
    bounding_box = BoundingBox(*grid_bounds)
    bounds = bounding_box.get_box_array()
    
    frame_norm_factor = (contrast_threshold > 0.0)
    
    if not img.flags['F_CONTIGUOUS']:
        img = np.array(img, order='F')        
        
    [frames, descs] =  _vlfeat.vl_dsift(img, step_size, bounds, int(descriptor_size/4), True, 1.5, verbose, frame_norm_factor)
    
    if frame_norm_factor:            
        f_contrast = frames[2, :]
        frames = frames[:2, :]
        f_highcontrast = f_contrast >= contrast_threshold
        if np.sum(f_highcontrast) == 0:
            raise ValueError('no descriptors could be found after contrast thresholding')
        frames = frames[:, f_highcontrast]
        descs = descs[:, f_highcontrast]
         
    return frames.T, descs.T

def normalize_image(image, verbose):
    '''
    Converts the image to gray scale float32, [0..1], 
    F continuous (column major order)
    ATTENTION if the maximal value > 1.0 assume original range [0..255]
              if the maximal value <=1.0 assume original range [0..1] 
    @param image: a numpy ndarray (NxMx1 or NxMx3) 
    '''
    image = np.array(image, dtype='float32', order='F')  # 'F' = column-major order!
    
    if image.ndim == 3:
        image = 0.2989 * image[:, :, 0] + 0.5870 * image[:, :, 1] + 0.1140 * image[:, :, 2]
    
    if image.ndim != 2:
        raise ValueError('Unsupported number of dimensions in image!')
        
    image_vmax = image.max()
    if image_vmax > 1.0:
        if verbose:
            print 'ATTENTION: Normalizing image from [0..255] to [0..1]'
            print '  --> estimated [0..255] due to maximal intensity value: %.2f' % image_vmax
        image = image / 255.0
        
    return image

class BoundingBox():
    def __init__(self, row_min, col_min, row_max, col_max, annotation=None):
        self.__annotation = annotation
        self.__row_min = row_min
        self.__row_max = row_max
        self.__col_min = col_min
        self.__col_max = col_max
        
    def get_box_array(self):
        box_array = [self.__row_min, self.__col_min, self.__row_max, self.__col_max]
        return np.array(box_array, 'f')