'''
@author: Sebastian Sudholt
@organization: TU Dortmund University, Dortmund, Germany
@contact: sebastian.sudholt@tu-dortmund.de
'''

import numpy as np

def map_from_pairwise_distances(pairwise_distances, word_list):
    '''
    Calculates the mAP from pairwise distances of a certain representation   
    @param pairwise_distances: (ndarray)
        n x n array indicating the pairwise distances of the global word image
        representations from the word list
    @param word_list: (list)
        list of tuples of quantized SIFT descriptors, normalized key points and
        transcription of the individual word images
    @return: returns the mean Average Precision
    '''
    listOfWordStrings = [word[0] for word in word_list]
    inds = np.argsort(pairwise_distances, axis=1)
    retr_mat = np.tile(listOfWordStrings, (len(listOfWordStrings),1))
    
    row_selector = np.transpose(np.tile(np.arange(len(word_list)), (len(word_list),1)))
    retr_mat = retr_mat[row_selector, inds]
    words = np.array(listOfWordStrings, ndmin=2).T
    
    # calculate the relevance matrix
    rel_matrix = retr_mat == words
    
    # as the query is included in the test up until now,
    # we need to remove it (the query will always be the
    # best retrieval result with distance 0)
    # -> drop the first element
    rel_matrix = rel_matrix[:,1:]
    
    map_calc = MeanAveragePrecision()
    avg_precs = np.array([map_calc.average_precision(row, n_relevant) 
                          for row, n_relevant in zip(rel_matrix, np.sum(rel_matrix, axis=1))])#, ndmin=2)
#     avg_precs = avg_precs.flatten()
    
    # queries that do not exist in the database yield an AP of 0
    # these queries are not considered in this benchmark
    # -> drop them
    mAP = np.mean(avg_precs[avg_precs > 0])
    return mAP

class IterativeMean(object):
    '''
    Class for iteratively computing a mean. With every new value (@see: _add_value)
    the mean will be updated 
    '''
    
    def __init__(self, mean_init=0.0):
        '''
        Constructor
        '''        
        self.__mean = mean_init
        self.__N = 0.0
        
    def add_value(self, value):
        '''
        Updates the mean with respect to value
        @param value: The value that will be incorporated in the mean
        '''  
        self.__mean = (self.__N / (self.__N + 1)) * self.__mean + (1.0 / (self.__N + 1)) * value
        self.__N += 1
        
    def get_mean(self):
        '''
        @return: The current mean 
        '''
        return self.__mean
    
    def reset(self):
        '''
        Resets the iterative mean calculation
        '''
        self.__mean = 0.0
        self.__N = 0.0         

class MeanAveragePrecision(IterativeMean):
    '''
    Computes average precision values and iteratively updates their mean
    @see: average_precision, get_mean, reset
    '''
    def __init__(self):
        '''
        Constructor
        '''
        super(MeanAveragePrecision, self).__init__()
        
    def average_precision(self, ret_vec_relevance, n_relevance):
        '''
        Computes the average precision and updates the mean average precision
        @param ret_vec_relevance: A 1-D numpy array containing ground truth (gt) 
            relevance values
        @return: The average precision for the given relevance vector
        '''
        ret_vec_cumsum = np.cumsum(ret_vec_relevance, dtype=float)
        ret_vec_range = np.arange(1, ret_vec_relevance.size + 1)
        ret_vec_precision = ret_vec_cumsum / ret_vec_range

        if n_relevance > 0:
            ret_vec_ap = (ret_vec_precision * ret_vec_relevance).sum() / n_relevance
        else:
            ret_vec_ap = 0.0
        
        super(MeanAveragePrecision, self).add_value(ret_vec_ap)        
        return ret_vec_ap

