'''
@author: Sebastian Sudholt
@organization: TU Dortmund University, Dortmund, Germany
@contact: sebastian.sudholt@tu-dortmund.de
'''

import numpy as np
from itertools import product

def extract_spatial_pyramid(quantized_descriptors, normalized_keypoint_locations, 
                            partitions, visual_vocabulary_size, power_norm_alpha=1,
                            reweighting_strategy='squared_partitions'):
    '''
    Extracts a spatial pyramid from the given quantized descriptors and the
    corresponding normalized key point locations.
    The partitions variable specifies the layout of the spatial pyramid
    and should be a tuple of tuples, where each sub-tuple represents
    the partitioning of a spatial pyramid level in height and width
    
    sample partition: ((2,3),(2,9))
    is a two-layer pyramid with 6 cells in the first layer
    (vertical axis is subdivided into 2 slices and the
    horizontal axis is subdivided into 3 slices)
    and 18 in the second. This setup was used in Aldavert2013-IVA
    
    @param quantized_descriptors: 
        (ndarray) one dimensional matrix containing the quantized
        descriptors
    @param normalized_keypoint_locations:
        (ndarray) two-dimensional matrix containing the relative key point
        locations of the corresponding quantized descriptors
        Note: the first element of each key point is the x-coordinate
        and the second is the y-coordinate. This way key points as generated
        by vlfeat can be used directly.
    @param partitions:
        (tuple of tuples) partitioning layout for the spatial pyramid as
        described above
    @param visual_vocabulary_size: 
        (int) the number of visual words used in creating the quantized
        descriptors
    @param power_norm_alpha: 
        (float) the scaling factor used in power normalization
    @param reweighting_strategy:
        (str) each level in the spatial pyramid has to be reweighted as not
        all quantized descriptors are used. Possible values are:
        squared_partitions: each cell is multiplied by the square of
                            cells on the corresponding level
        partitions: each cell is multiplied by the number of cells
                    on the corresponding level
    @return: the spatial pyramid representation of the supplied data
                                
    '''
    # some checks
    if not quantized_descriptors.shape[0] == normalized_keypoint_locations.shape[0]:
        raise ValueError('The number of key points and quantized descriptors does not match')
    if reweighting_strategy not in ['squared_partitions', 'partitions']:
        raise ValueError('The reweighting strategy %s is unknown' % (reweighting_strategy))
    
    bof_hist_list = []
    for level_idx, cur_level_partition in enumerate(partitions):
        if not (len(cur_level_partition) == 2):
            raise ValueError('Invalid partition on level %i' % (level_idx))
        for cur_bin in product(xrange(cur_level_partition[0]), xrange(cur_level_partition[1])):            
            # the key points are represented in (x,y) coordinates -> the indices have to be flipped
            upper_left = np.array((float(cur_bin[1])/float(cur_level_partition[1]), float(cur_bin[0])/float(cur_level_partition[0])))
            lower_right = np.array((float(cur_bin[1]+1)/float(cur_level_partition[1]), float(cur_bin[0]+1)/float(cur_level_partition[0])))
            bool_vals_XY = np.logical_and(normalized_keypoint_locations > upper_left, 
                                          normalized_keypoint_locations <= lower_right)
            indicesBin = np.logical_and(bool_vals_XY[:,0], bool_vals_XY[:,1])
            
            # extract the bof histogram
            bof_hist = np.bincount(quantized_descriptors[indicesBin].astype(np.int32).flatten(), 
                                   minlength=visual_vocabulary_size)
            
            # reweighting
            if reweighting_strategy == 'partitions':
                # right now we use a weighting scheme similar to the one used in Rusinol2015-ESF:
                # multiply each cell with the number of total cells on this level
                bof_hist *= cur_level_partition[0]*cur_level_partition[1]
            elif reweighting_strategy == 'squared_partitions':
                # multiply by the square of cells on the current layer
                bof_hist *= (cur_level_partition[0]*cur_level_partition[1])**2
            
            bof_hist_list.append(bof_hist)
    # build the pyramid
    pyramid = np.hstack(bof_hist_list)
    
    # power normalize if necessary
    pyramid = np.power(pyramid, power_norm_alpha)
    return pyramid