'''
@author: Sebastian Sudholt
@organization: TU Dortmund University, Dortmund, Germany
@contact: sebastian.sudholt@tu-dortmund.de
'''

import os
import numpy as np
import math
import argparse

from vlfeat_sift_bindings import vl_dsift

def load_word_images(pages_dir, ground_truth_dir, descriptor_size, step_size, grid_offset, contrast_threshold):
    '''
    This function loads the document images and the corresponding ground truth
    for a document image collection and computes SIFT descriptors in a dense
    grid for each word image.
    
    @param pages_dir: (str)
        absolute path to the document image root directory (files must be PNG)
    @param ground_truth_dir: (str)
        absolute path to the ground truth root directory
    @param descriptor_size: (int)
        width of the desired SIFT descriptor to be extracted (as the descriptor
        is square it is also its height)
    @param step_size: (int)
        the step size to be used when extracting the dense grid
    @param grid_offset: (tuple)
        the upper left corner of the dense grid in pixel coordinates
    @param contrast_threshold: (float)
        threshold for gradient magnitude in computing the SIFT descriptors
    '''
    word_list = []
    gt_files = sorted(os.listdir(ground_truth_dir))
    page_files = sorted(os.listdir(pages_dir))
    for idx, (gt_filename, page_filename) in enumerate(zip(gt_files, page_files)):
        print('\t[ %*d / %d ] Processing page %s... ' % (int(math.log10(len(page_files)))+1,
                                                         idx+1, len(page_files), page_filename)),
        keypoints, descriptors = vl_dsift.dense_features(image_filepath=os.path.join(pages_dir, page_filename), 
                                                         descriptor_size=descriptor_size, step_size=step_size, 
                                                         grid_offset=grid_offset, contrast_threshold=contrast_threshold, verbose=False)
        with open(os.path.join(ground_truth_dir, gt_filename), 'r') as gt_file:
            for line in gt_file:
                line_parts = line.split(' ')
                coordinate_list = [int(x) for x in line_parts[:4]]
                upper_left = np.array(coordinate_list[:2])
                lower_right = np.array(coordinate_list[2:])
                
                # extract descriptors and keypoints for each word
                bool_vals_XY = np.logical_and(keypoints > upper_left, 
                                              keypoints < lower_right)                                
                indices = np.logical_and(bool_vals_XY[:,0],bool_vals_XY[:,1])
                if np.sum(indices) == 0:
                    print 'no descriptors for %s, ' % (line_parts[4]),
                    continue
                word_normalized_keypoints = keypoints[indices] - upper_left
                word_normalized_keypoints /= lower_right-upper_left
                word_descriptors = descriptors[indices]
                
                word_list.append((line_parts[4], word_normalized_keypoints, word_descriptors))
        print('[ Done ]')
    return word_list

def spatial_pyramid_parser(sp_string):
    '''
    Parses a spatial pyramid string representation into a tuple of tuples.
    @param sp_string: (str)
        the string representation of the spatial pyramid
    '''
    try:
        levels = sp_string.split('),(')
        spatial_pyramid_tuple = []
        for level in levels:
            level = level.replace('(','').replace(')','')
            level_partition = tuple([int(x) for x in level.split(',')])
            if len(level_partition) != 2:
                raise argparse.ArgumentError()
            spatial_pyramid_tuple.append(level_partition)
        bla = tuple(spatial_pyramid_tuple)
        return bla
    except:
        raise argparse.ArgumentError('The spatial pyramid layout is incorrect')
